/* $Id: completion.c,v 1.21 2004/04/09 21:52:33 deletek Exp $ */

/*
 *  (C) Copyright 2003 Wojtek Kaniewski <wojtekka@irc.pl>
 *                2004 Piotr Kupisiewicz <deli@rzepaknet.us>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <dirent.h>
#ifndef HAVE_SCANDIR
#  include "compat/scandir.h"
#endif

#include <ekg/dynstuff.h>
#include <ekg/xmalloc.h>
#include <ekg/commands.h>
#include <ekg/stuff.h>
#include <ekg/userlist.h>
#include <ekg/vars.h>

#include "old.h"

/* nadpisujemy funkcj strncasecmp() odpowiednikiem z obsug polskich znakw */
#define strncasecmp(x...) strncasecmp_pl(x)

static char **completions = NULL;	/* lista dopenie */

static void dcc_generator(const char *text, int len)
{
	const char *words[] = { "close", "get", "send", "list", "rsend", "rvoice", "voice", NULL };
	int i;

	for (i = 0; words[i]; i++)
		if (!strncasecmp(text, words[i], len))
			array_add(&completions, xstrdup(words[i]));
}

static void command_generator(const char *text, int len)
{
	const char *slash = "", *dash = "";
	list_t l;

	if (*text == '/') {
		slash = "/";
		text++;
		len--;
	}

	if (*text == '^') {
		dash = "^";
		text++;
		len--;
	}

	if (window_current->target)
		slash = "/";
			
	for (l = commands; l; l = l->next) {
		command_t *c = l->data;
		char *without_sess_id = strchr(c->name, ':');

		if (!strncasecmp(text, c->name, len) && !array_item_contains(completions, c->name, 1))
			array_add(&completions, saprintf("%s%s%s", slash, dash, c->name));
		else if (without_sess_id && !array_item_contains(completions, without_sess_id + 1, 1) && !strncasecmp(text, without_sess_id + 1, len)) 
			array_add(&completions, saprintf("%s%s%s", slash, dash, without_sess_id + 1));
	}
}

static void events_generator(const char *text, int len)
{
#if 0
	int i;
	const char *tmp = NULL;
	char *pre = NULL;

	if ((tmp = strrchr(text, '|')) || (tmp = strrchr(text, ','))) {
		char *foo;

		pre = xstrdup(text);
		foo = strrchr(pre, *tmp);
		*(foo + 1) = 0;

		len -= tmp - text + 1;
		tmp = tmp + 1;
	} else
		tmp = text;

	for (i = 0; event_labels[i].name; i++)
		if (!strncasecmp(tmp, event_labels[i].name, len))
			array_add(&completions, ((tmp == text) ? xstrdup(event_labels[i].name) : saprintf("%s%s", pre, event_labels[i].name)));
#endif
}

static void ignorelevels_generator(const char *text, int len)
{
	int i;
	const char *tmp = NULL;
	char *pre = NULL;

	if ((tmp = strrchr(text, '|')) || (tmp = strrchr(text, ','))) {
		char *foo;

		pre = xstrdup(text);
		foo = strrchr(pre, *tmp);
		*(foo + 1) = 0;

		len -= tmp - text + 1;
		tmp = tmp + 1;
	} else
		tmp = text;

	for (i = 0; ignore_labels[i].name; i++)
		if (!strncasecmp(tmp, ignore_labels[i].name, len))
			array_add(&completions, ((tmp == text) ? xstrdup(ignore_labels[i].name) : saprintf("%s%s", pre, ignore_labels[i].name)));
}

static void unknown_uin_generator(const char *text, int len)
{
	int i;
	
	for (i = 0; i < send_nicks_count; i++) {
		if (send_nicks[i] && xstrchr(send_nicks[i], ':') && xisdigit(xstrchr(send_nicks[i], ':')[1]) && !strncasecmp(text, send_nicks[i], len))
			if (!array_contains(completions, send_nicks[i], 0))
				array_add(&completions, xstrdup(send_nicks[i]));
	}
}

static void known_uin_generator(const char *text, int len)
{
	int done = 0;
	list_t l;
	session_t *s;

	if (!session_current)
		return;

	s  = session_current;

	for (l = s->userlist; l; l = l->next) {
		userlist_t *u = l->data;
		if (u->nickname && !strncasecmp(text, u->nickname, len)) {
			array_add(&completions, xstrdup(u->nickname));
			done = 1;
		}
	}

	for (l = s->userlist; l; l = l->next) {
		userlist_t *u = l->data;

		if (!done && !strncasecmp(text, u->uid, len))
			array_add(&completions, xstrdup(u->uid));
	}
	
	for (l = conferences; l; l = l->next) {
		struct conference *c = l->data;

		if (!strncasecmp(text, c->name, len))
			array_add(&completions, xstrdup(c->name));
	}

	unknown_uin_generator(text, len);
}

static void variable_generator(const char *text, int len)
{
	list_t l;

	for (l = variables; l; l = l->next) {
		variable_t *v = l->data;

		if (v->type == VAR_FOREIGN)
			continue;

		if (*text == '-') {
			if (!strncasecmp(text + 1, v->name, len - 1))
				array_add(&completions, saprintf("-%s", v->name));
		} else {
			if (!strncasecmp(text, v->name, len))
				array_add(&completions, xstrdup(v->name));
		}
	}
}

static void ignored_uin_generator(const char *text, int len)
{
        session_t *s;
	list_t l;

        if (!session_current)
                return;

        s  = session_current;

	for (l = s->userlist; l; l = l->next) {
		userlist_t *u = l->data;

		if (!ignored_check(s, u->uid))
			continue;

		if (!u->nickname) {
			if (!strncasecmp(text, u->uid, len))
				array_add(&completions, xstrdup(u->uid));
		} else {
			if (u->nickname && !strncasecmp(text, u->nickname, len))
				array_add(&completions, xstrdup(u->nickname));
		}
	}
}

static void blocked_uin_generator(const char *text, int len)
{
        session_t *s;
	list_t l;

        if (!session_current)
                return;

        s  = session_current;

	for (l = s->userlist; l; l = l->next) {
		userlist_t *u = l->data;

		if (!ekg_group_member(u, "__blocked"))
			continue;

		if (!u->nickname) {
			if (!strncasecmp(text, u->uid, len))
				array_add(&completions, xstrdup(u->uid));
		} else {
			if (u->nickname && !strncasecmp(text, u->nickname, len))
				array_add(&completions, xstrdup(u->nickname));
		}
	}
}

static void empty_generator(const char *text, int len)
{

}

void file_generator(const char *text, int len)
{
	struct dirent **namelist = NULL;
	char *dname, *tmp;
	const char *fname;
	int count, i;

	/* `dname' zawiera nazw katalogu z koczcym znakiem `/', albo
	 * NULL, jeli w dopenianym tekcie nie ma cieki. */

	dname = xstrdup(text);

	if ((tmp = strrchr(dname, '/'))) {
		tmp++;
		*tmp = 0;
	} else
		dname = NULL;

	/* `fname' zawiera nazw szukanego pliku */

	fname = strrchr(text, '/');

	if (fname)
		fname++;
	else
		fname = text;

again:
	/* zbierzmy list plikw w danym katalogu */
	
	count = scandir((dname) ? dname : ".", &namelist, NULL, alphasort);

	debug("dname=\"%s\", fname=\"%s\", count=%d\n", (dname) ? dname : "(null)", (fname) ? fname : "(null)", count);

	for (i = 0; i < count; i++) {
		char *name = namelist[i]->d_name, *tmp = saprintf("%s%s", (dname) ? dname : "", name);
		struct stat st;
		int isdir = 0;

		if (!stat(tmp, &st))
			isdir = S_ISDIR(st.st_mode);

		xfree(tmp);

		if (!strcmp(name, ".")) {
			xfree(namelist[i]);
			continue;
		}

		/* jeli mamy `..', sprawd czy katalog skada si z
		 * `../../../' lub czego takiego. */
		
		if (!strcmp(name, "..")) {
			const char *p;
			int omit = 0;

			for (p = dname; p && *p; p++) {
				if (*p != '.' && *p != '/') {
					omit = 1;
					break;
				}
			}

			if (omit) {
				xfree(namelist[i]);
				continue;
			}
		}
		
		if (!strncmp(name, fname, strlen(fname))) {
			name = saprintf("%s%s%s", (dname) ? dname : "", name, (isdir) ? "/" : "");
			array_add(&completions, name);
		}

		xfree(namelist[i]);
        }

	/* jeli w dopenieniach wyldowa tylko jeden wpis i jest katalogiem
	 * to wejd do niego i szukaj jeszcze raz */

	if (array_count(completions) == 1 && strlen(completions[0]) > 0 && completions[0][strlen(completions[0]) - 1] == '/') {
		xfree(dname);
		dname = xstrdup(completions[0]);
		fname = "";
		array_free(completions);
		completions = NULL;

		goto again;
	}

	xfree(dname);
	xfree(namelist);
}

static void python_generator(const char *text, int len)
{
	const char *words[] = { "load", "unload", "run", "exec", "list", NULL };
	int i;

	for (i = 0; words[i]; i++)
		if (!strncasecmp(text, words[i], len))
			array_add(&completions, xstrdup(words[i]));
}

static void window_generator(const char *text, int len)
{
	const char *words[] = { "new", "kill", "move", "next", "resize", "prev", "switch", "clear", "refresh", "list", "active", "last", NULL };
	int i;

	for (i = 0; words[i]; i++)
		if (!strncasecmp(text, words[i], len))
			array_add(&completions, xstrdup(words[i]));
}

static void sessions_generator(const char *text, int len)
{
        list_t l;

        for (l = sessions; l; l = l->next) {
                session_t *v = l->data;
                if (*text == '-') {
                        if (!strncasecmp(text + 1, v->uid, len - 1))
                                array_add_check(&completions, saprintf("-%s", v->uid), 1);
                } else {
                        if (!strncasecmp(text, v->uid, len))
                                array_add_check(&completions, xstrdup(v->uid), 1);
                }
        }
}

static void sessions_var_generator(const char *text, int len)
{
        list_t l;
	int i;
        const char *words[] = { "uid", "password", "auto_connect", "alias", "server", NULL };

        for (i = 0; words[i]; i++) {
		if(*text == '-') {
	        	if (!strncasecmp(text + 1, words[i], len - 1))
        	        	array_add_check(&completions, saprintf("-%s", words[i]), 1);
		} else {
			if (!strncasecmp(text, words[i], len))
                                array_add_check(&completions, xstrdup(words[i]), 1);
		}
	}
 
	for (l = sessions; l; l = l->next) {
                session_t *v = l->data;

		for(i = 0; v->params && v->params[i]; i++) {
                	if (*text == '-') {
                        	if (!strncasecmp(text + 1, v->params[i]->key, len - 1)) 
                                	array_add_check(&completions, saprintf("-%s", v->params[i]->key), 1);
	                } else {
        	                if (!strncasecmp(text, v->params[i]->key, len))
                	                array_add_check(&completions, xstrdup(v->params[i]->key), 1);
 	               }
		}
        }
}

static void sessions_all_generator(const char *text, int len)
{
	sessions_var_generator(text, len);
	sessions_generator(text, len);
}

static struct {
	char ch;
	void (*generate)(const char *text, int len);
} generators[] = {
	{ 'u', known_uin_generator },
	{ 'U', unknown_uin_generator },
	{ 'c', command_generator },
	{ 'x', empty_generator },
	{ 'i', ignored_uin_generator },
	{ 'b', blocked_uin_generator },
	{ 'v', variable_generator },
	{ 'd', dcc_generator },
	{ 'p', python_generator },
	{ 'w', window_generator },
	{ 'f', file_generator },
	{ 'e', events_generator },
        { 's', sessions_generator },
	{ 'S', sessions_var_generator },
        { 'A', sessions_all_generator },
	{ 'I', ignorelevels_generator },
	{ 0, NULL }
};

/*
 * ncurses_complete()
 *
 * funkcja obsugujca dopenianie klawiszem tab.
 * Dziaanie:
 * - Wprowadzona linia dzielona jest na wyrazy (uwzgldniajc przecinki i znaki cudzyslowia)
 * - nastpnie znaki separacji znajdujce si midzy tymi wyrazami wrzucane s do tablicy separators
 * - dalej sprawdzane jest za pomoc zmiennej word_current (okrelajcej aktualny wyraz bez uwzgldnienia
 *   przecinkw - po to, aby wiedzie czy w przypadku np funkcji /query ma by szukane dopenienie 
 * - zmienna word odpowiada za aktualny wyraz (*z* uwzgldnieniem przecinkw)
 * - words - tablica zawieraj wszystkie wyrazy
 * - gdy jest to moliwe szukane jest dopenienie 
 * - gdy dopenie jest wicej ni jedno (count > 1) wywietlamy tylko "wspln" cz wszystkich dopenie
 *   np ,,que'' w przypadku funkcji /query i /queue
 * - gdy dopenienie jest tylko jedno wywietlamy owo dopenienie
 * - przy wywietlaniu dopenienia caa linijka konstruowana jest od nowa, poniewa nie wiadomo w ktrym miejscu
 *   podany wyraz ma zosta "wsadzony", std konieczna jest tablica separatorw, tablica wszystkich wyrazw itd ...
 */
void ncurses_complete(int *line_start, int *line_index, char *line)
{
	char *start, *cmd, **words, *separators;
	int i, count, word, j, words_count, word_current;

	start = xmalloc(strlen(line) + 1);
	count = 0;
	/* 
	 * jeli uzbierano ju co to prbujemy wywietli wszystkie moliwoci 
	 */
	if (completions) {
		int maxlen = 0, cols, rows;
		char *tmp;

		for (i = 0; completions[i]; i++)
			if (strlen(completions[i]) + 2 > maxlen)
				maxlen = strlen(completions[i]) + 2;

		cols = (window_current->width - 6) / maxlen;
		if (cols == 0)
			cols = 1;

		rows = array_count(completions) / cols + 1;

		tmp = xmalloc(cols * maxlen + 2);

		for (i = 0; i < rows; i++) {
			int j;

			strcpy(tmp, "");

			for (j = 0; j < cols; j++) {
				int cell = j * rows + i;

				if (cell < array_count(completions)) {
					int k;

					strcat(tmp, completions[cell]);

					for (k = 0; k < maxlen - strlen(completions[cell]); k++)
						strcat(tmp, " ");
				}
			}

			if (strcmp(tmp, "")) {
				print("none", tmp);
			}
		}

		xfree(tmp);
		xfree(start);
		return;
	}

	/* zerujemy co mamy */
	words = NULL;

	/* podziel (uwzgldniajc cudzysowia)*/
	for (i = 0; i < strlen(line); i++) {
		if(line[i] == '"')
			for(j = 0,  i++; i < strlen(line) && line[i] != '"'; i++, j++)
				start[j] = line[i];
		else
			for(j = 0; i < strlen(line) && !xisspace(line[i]) && line[i] != ','; j++, i++)
				start[j] = line[i];
		start[j] = '\0';
		/* "przewijamy" wiksz ilo spacji */
		for(i++; i < strlen(line) && (xisspace(line[i]) || line[i] == ','); i++);
		i--;
		array_add(&words, saprintf("%s", start));
	}
	
	/* jeeli ostatnie znaki to spacja, albo przecinek to trzeba doda jeszcze pusty wyraz do words */
	if (strlen(line) > 1 && (line[strlen(line) - 1] == ' ' || line[strlen(line) - 1] == ','))
		array_add(&words, xstrdup(""));

/*	 for(i = 0; i < array_count(words); i++)
		debug("words[i = %d] = \"%s\"\n", i, words[i]);     */

	/* inicjujemy pamic dla separators */
	if (words != NULL)
		separators = xmalloc(array_count(words));
	else
		separators = NULL;

	/* sprawd, gdzie jestemy (uwzgdniajc cudzysowia) i dodaj separatory*/
	for (word = 0, i = 0; i < strlen(line); i++, word++) {
		if(line[i] == '"')  {
			for(j = 0, i++; i < strlen(line) && line[i] != '"'; j++, i++)
				start[j] = line[i];
		} else {
			for(j = 0; i < strlen(line) && !xisspace(line[i]) && line[i] != ','; j++, i++)
				start[j] = line[i];
		}
		/* "przewijamy */
		for(i++; i < strlen(line) && (xisspace(line[i]) || line[i] == ','); i++);
		/* ustawiamy znak koca */
		start[j] = '\0';
		/* jeeli to koniec linii, to koczymy t zabaw */
		if(i >= strlen(line))
	    		break;
		/* obniamy licznik o 1, eby wszystko byo okey, po "przewijaniu" */
		i--;
		/* hmm, jestemy ju na wyrazie wskazywany przez kursor ? */
                if(i >= *line_index)
            		break;
	}

	/* dodajmy separatory - pewne rzeczy podobne do ptli powyej */
	for (i = 0, j = 0; i < strlen(line); i++, j++) {
		if(line[i] == '"')  {
			for(i++; i < strlen(line) && line[i] != '"'; i++);
			if(i < strlen(line)) 
				separators[j] = line[i + 1];
		} else {
			for(; i < strlen(line) && !xisspace(line[i]) && line[i] != ','; i++);
			separators[j] = line[i];
		}

		for(i++; i < strlen(line) && (xisspace(line[i]) || line[i] == ','); i++);
		i--;
	}

	if (separators)
		separators[j] = '\0'; // koniec ciagu
	
	/* aktualny wyraz bez uwzgledniania przecinkow */
	for (i = 0, words_count = 0, word_current = 0; i < strlen(line); i++, words_count++) {
		for(; i < strlen(line) && !xisspace(line[i]); i++)
			if(line[i] == '"') 
				for(i++; i < strlen(line) && line[i] != '"'; i++);
		for(i++; i < strlen(line) && xisspace(line[i]); i++);
		if(i >= strlen(line))
			word_current = words_count + 1;
		i--;
                /* hmm, jestemy ju na wyrazie wskazywany przez kursor ? */
                if(i >= *line_index)
                        break;
	}

	/* trzeba pododawa troch do licznikw w spefycicznych (patrz warunki) sytuacjach */
	if((xisspace(line[strlen(line) - 1]) || line[strlen(line) - 1] == ',') && word + 1== array_count(words) -1 )
		word++;
	if(xisspace(line[strlen(line) - 1]) && words_count == word_current)
		word_current++;
	if(xisspace(line[strlen(line) - 1]))
		words_count++;

/*	debug("word = %d\n", word);
	debug("start = \"%s\"\n", start);
	debug("words_count = %d\n", words_count);	 

	 for(i = 0; i < strlen(separators); i++)
		debug("separators[i = %d] = \"%c\"\n", i, separators[i]);   */

	cmd = saprintf("/%s ", (config_tab_command) ? config_tab_command : "chat");

	/* nietypowe dopenienie nickw przy rozmowach */
	if (!strcmp(line, "") || (!strncasecmp(line, cmd, strlen(cmd)) && word == 2 && send_nicks_count > 0) || (!strcasecmp(line, cmd) && send_nicks_count > 0)) {
		if (send_nicks_index >= send_nicks_count)
			send_nicks_index = 0;

		if (send_nicks_count) {
			char *nick = send_nicks[send_nicks_index++];

			snprintf(line, LINE_MAXLEN, (strchr(nick, ' ')) ? "%s\"%s\" " : "%s%s ", cmd, nick);
		} else
			snprintf(line, LINE_MAXLEN, "%s", cmd);
		*line_start = 0;
		*line_index = strlen(line);

                array_free(completions);
                array_free(words);
		xfree(start);
		xfree(separators);
		xfree(cmd);
		return;
	}
	xfree(cmd);

	/* pocztek komendy? */
	if (word == 0)
		command_generator(start, strlen(start));
	else {
		char *params = NULL;
		int abbrs = 0, i;
		list_t l;

		for (l = commands; l; l = l->next) {
			command_t *c = l->data;
			char *name = (strchr(c->name, ':')) ? strchr(c->name, ':') + 1 : c->name;
			int len = strlen(name);
			char *cmd = (line[0] == '/') ? line + 1 : line;
			
			if (!strncasecmp(name, cmd, len) && xisspace(cmd[len])) {
				params = c->params;
				abbrs = 1;
				break;
			}

			for (len = 0; cmd[len] && cmd[len] != ' '; len++);

			if (!strncasecmp(name, cmd, len)) {
				params = c->params;
				abbrs++;
			} else
				if (params && abbrs == 1)
					break;
		}
		
		if (params && abbrs == 1) {
			for (i = 0; generators[i].ch; i++) {
				if (generators[i].ch == params[word_current - 2]) {
					int j;

					generators[i].generate(words[word], strlen(words[word]));

					for (j = 0; completions && completions[j]; j++) {
						string_t s;
						const char *p;

						if (!strchr(completions[j], '"') && !strchr(completions[j], '\\') && !strchr(completions[j], ' '))
							continue;

						s = string_init("\"");

						for (p = completions[j]; *p; p++) {
							if (!strchr("\"\\", *p))
								string_append_c(s, *p);
							else {
								string_append_c(s, '\\');
								string_append_c(s, *p);
							}
						}
						string_append_c(s, '\"');

						xfree(completions[j]);
						completions[j] = string_free(s, 0);
					}
					break;
				}
			}
		}
	}

	count = array_count(completions);
	
	/* 
	 * jeli jest tylko jedna molwio na dopenienie to drukujemy co mamy,
	 * ewentualnie bierzemy cz wyrazw w cudzysowia ...
	 * i uwaamy oczywicie na \001 (patrz funkcje wyej
	 */
	if (count == 1) {
		line[0] = '\0';		
		for(i = 0; i < array_count(words); i++) {
			if(i == word) {
				if(strchr(completions[0],  '\001')) {
					if(completions[0][0] == '"')
						strncat(line, completions[0] + 2, strlen(completions[0]) - 2 - 1 );
					else
						strncat(line, completions[0] + 1, strlen(completions[0]) - 1);
				} else
			    		strcat(line, completions[0]);
				*line_index = strlen(line) + 1;
			} else {
				if(strchr(words[i], ' '))
					strcat(line, saprintf("\"%s\"", words[i]));
				else
					strcat(line, words[i]);
			}
			if((i == array_count(words) - 1 && line[strlen(line) - 1] != ' ' ))
				strcat(line, " ");
			else if (line[strlen(line) - 1] != ' ')
                                strcat(line, saprintf("%c", separators[i]));
		}
		array_free(completions);
		completions = NULL;
	}

	/*
	 * gdy jest wicej moliwoci to robimy podobnie jak wyej tyle, e czasem
	 * trzeba uy cudzysowia tylko z jednej storny, no i trzeba dopeni do pewnego miejsca
	 * w sumie proste rzeczy, ale jak wida jest troch opcji ...
	 */
	if (count > 1) {
		int common = 0;
		int tmp = 0;
		int quotes = 0;

	    	/* for(i = 0; completions[i]; i++)
                	debug("completions[i] = %s\n", completions[i]); */
		/*
		 * moe nie za adne programowanie, ale skuteczne i w sumie jedyne w 100% speniajce	
	 	 * wymagania dopeniania (uwzgldnianie cudzywsowiw itp...)
		 */
		for(i=1, j = 0; ; i++, common++) {
			for(j=0; j < count; j++) {
//				if (!completions[j][i])
//					break;
				if(completions[j][0] == '"') 
					quotes = 1;
				if(completions[j][0] == '"' && completions[0][0] != '"')
					tmp = strncasecmp(completions[0], completions[j] + 1, i);
				else if(completions[0][0] == '"' && completions[j][0] != '"')
					tmp = strncasecmp(completions[0] + 1, completions[j], i);
				else
					tmp = strncasecmp(completions[0], completions[j], i);
				 /* debug("strncasecmp(\"%s\", \"%s\", %d) = %d\n", completions[0], completions[j], i, strncasecmp(completions[0], completions[j], i));  */
				if (tmp)
					break;
                        }
			if (tmp)
				break;
		}
		
		/* debug("common :%d\n", common); */

		if (strlen(line) + common < LINE_MAXLEN) {
		
			line[0] = '\0';
			for(i = 0; i < array_count(words); i++) {
				if(i == word) {
					if(quotes == 1 && completions[0][0] != '"') 
						strcat(line, "\"");

					if(completions[0][0] == '"')
						common++;
						
					if(completions[0][common - 1] == '"')
						common--;

					strncat(line, completions[0], common);
					*line_index = strlen(line);
				} else {
					if(strrchr(words[i], ' '))
						strcat(line, saprintf("\"%s\"", words[i]));
					else
						strcat(line, words[i]);
				}

				if(separators[i]) {
					strcat(line, saprintf("%c", separators[i]));
				}
			}
		}
	}

	array_free(words);
	xfree(start);
	xfree(separators);
	return;
}
void ncurses_complete_clear()
{
	array_free(completions);
	completions = NULL;
}

